' Let it snow 2013 by Krischan
' ----------------------------
'
' Image from www.bunnybua.com, redesigned by Krischan
'
' Xmas Music is a converted free MIDI - resampled with Renoise by Krischan
' License: http://creativecommons.org/licenses/by-nc-sa/3.0/
'
' Firework Music is a converted free MIDI - resampled with Renoise by Krischan
' License: http://creativecommons.org/licenses/by-nd/3.0/
'
' Rocket Sound FX from http://www.freesounds.org
' License: http://creativecommons.org/licenses/by/3.0/

SuperStrict

Import "santa.o"

Framework brl.GLMax2D
Import brl.WAVloader
Import brl.TGAloader
Import brl.Pixmap
Import brl.Random
Import brl.OGGloader
Import brl.DirectSoundAudio
Import brl.timer
Import brl.RamStream
Import brl.EventQueue
Import maxgui.Drivers
Import maxgui.ProxyGadgets

Incbin "images/xmas.tga"
Incbin "images/newyear.tga"

Incbin "music/Silent Night.ogg"
Incbin "music/Alla Hornpipe.ogg"

Incbin "particles/snow.tga"
Incbin "particles/particle1.tga"
Incbin "particles/particle2.tga"
Incbin "particles/particle3.tga"
Incbin "particles/particle4.tga"
Incbin "particles/particle5.tga"
Incbin "particles/particle6.tga"
Incbin "particles/particle7.tga"
Incbin "particles/particle8.tga"
Incbin "particles/rocket.tga"
Incbin "particles/bottle.tga"

Incbin "sound/explode1.ogg"
Incbin "sound/explode2.ogg"
Incbin "sound/explode3.ogg"
Incbin "sound/explode4.ogg"
Incbin "sound/explode5.ogg"
Incbin "sound/explode6.ogg"
Incbin "sound/explode7.ogg"
Incbin "sound/launch.ogg"

Global SW:Int = 1024			' window initial resolution width
Global SH:Int = 640				' window initial resolution height

Const Fullscreen:Int = True		' Windows or no border?
Const Maximize:Int = True		' Maximize to Desktop resolution?

Const rockets:Int = 25			' there must be at least 13 rockets / bottles
Const startflakes:Int = 0		' start amount of flakes
Const totalflakes:Int = 5000	' maximum amount of flakes
Const snowchance:Float = 25		' chance in % For New flakes
Const chanceflakes:Float = 25	' randomize amount of New flakes (1...x)
Const chancestick:Float = 0.9	' change that a flake sticks To the matte (here: 50%)

Global s:snowflake
Global fireworkcnt:Int
Global particlecnt:Int
Global mode:Int = 0
Global musicchannel:TChannel

Local timer:TTimer = CreateTimer(125)
Local init:Int = True

Type snowflake

	Global List:TList = CreateList()
	
	Method New()

		ListAddLast(List, Self)

	End Method
	
	Field x:Float, y:Float
	Field col:Int
	Field dirx:Float, diry:Float
	Field style:Float
	Field sinus:Float
	Field alpha:Float
	
End Type

Type firework

	Global List:TList = CreateList()
	
	Method New()

		ListAddLast(List, Self)

	End Method

	Method Remove()

		ListRemove(List, Self)

	End Method

	Field x:Float[5000], y:Float[5000]
	Field dirx:Float[5000], diry:Float[5000]
	Field size:Float[5000]
	Field grav:Float[5000]
	Field speed:Float[5000]
	Field particle:Int[5000]

	Field timer:Float
	Field starttimer:Int
	Field life:Int
	Field lifetimer:Int
	Field exploded:Int
	Field num:Int
	Field init:Int

End Type

' create window and canvas
Global window:TGadget
If Fullscreen Then

	window = CreateWindow("Let it snow", ClientWidth(Desktop()) / 2 - (SW / 2), ClientHeight(Desktop()) / 2 - (SH / 2), SW + 12, SH + 24, Null, Null)
	
Else

	window = CreateWindow("Let it snow", ClientWidth(Desktop()) / 2 - (SW / 2), ClientHeight(Desktop()) / 2 - (SH / 2), SW + 12, SH + 24, Null, WINDOW_TITLEBAR | WINDOW_RESIZABLE)

EndIf

If Maximize Then MaximizeWindow(window)


Global canvas:TGadget = CreateCanvas(0, 0, ClientWidth(window), ClientHeight(window), window, 0)
SetGadgetLayout canvas, 1, 1, 1, 1
SetGraphics CanvasGraphics(canvas)

' load media
Global explode:TSound[8]
Global particle:TImage[9]

For Local i:Int = 1 To 8

	particle[i] = LoadImage("incbin::particles/particle" + i + ".tga")

Next

For Local i:Int = 1 To 7

	explode[i] = LoadSound("incbin::sound/explode" + i + ".ogg")

Next

Global fly:TSound = LoadSound("incbin::sound/launch.ogg")
Global logo1:TPixmap = LoadPixmap("incbin::images/xmas.tga")
Global logo2:TPixmap = LoadPixmap("incbin::images/newyear.tga")
Global flake:TImage = LoadImage("incbin::particles/snow.tga")
Global rocket:TImage = LoadImage("incbin::particles/rocket.tga")
Global bottle:TImage = LoadImage("incbin::particles/bottle.tga")
Global image1:TPixmap = ResizePixmap(logo1, SW, SH)
Global image2:TPixmap = ResizePixmap(logo2, SW, SH)

Global music1:TSound = LoadSound("incbin::music/Silent Night.ogg", True)
Global music2:TSound = LoadSound("incbin::music/Alla Hornpipe.ogg", True)

SW = ClientWidth(window)
SH = ClientHeight(window)
image1 = ResizePixmap(logo1, SW, SH)
image2 = ResizePixmap(logo2, SW, SH)

' set Escape hotkey
SetHotKeyEvent(KEY_ESCAPE, Null, CreateEvent(EVENT_GADGETACTION, window, 0))
SetHotKeyEvent(KEY_SPACE, Null, CreateEvent(EVENT_GADGETACTION, window, 1))

' init snow
CreateSnow(startflakes)

SetGraphics CanvasGraphics(canvas)

InitFirework()

' ----------------------------------------------------------------------------
' Main Loop
' ----------------------------------------------------------------------------
While True

	particlecnt = 0

	ActivateGadget(canvas)

	WaitEvent()

	If EventSource() = timer Then

		If mode = 0 Then

			DrawPixmap(image1, 0, 0)
			SetGadgetText(window, "Let it snow " + UpdateSnow(s) + " flakes")

		Else

			DrawPixmap(image2, 0, 0)

			SetBlend(ALPHABLEND)
			SetAlpha(1.0)
			SetColor(255, 255, 255)
			SetRotation(0)

			For Local f:firework = EachIn firework.List

				If (f.y[1] < (SH - (SH * f.timer)) Or f.x[1] < 0 Or f.x[1] > SW) And f.exploded = 0 Then

					For Local i:Int = 1 To f.num

						f.x[i] = f.x[1]
						f.y[i] = f.y[1]
						f.grav[i] = Rnd(0.1, 0.5)
						f.speed[i] = Rnd(Rnd(1, 4))

						If Rnd(1) > 0.8 Then
							f.particle[i] = Rand(1, 7)
							f.size[i] = Rnd(4, 16)
						Else
							f.particle[i] = Rand(7, 8)
							f.size[i] = Rnd(4, 16)
						EndIf

					Next

					f.exploded = 1
					f.lifetimer = MilliSecs()
					SetChannelVolume(PlaySound(explode[Rand(1, 7)]), Rnd(0.5, 1))

				EndIf

				If f.exploded = 1 Then

					SetBlend(LIGHTBLEND)
					SetRotation(0)

					For Local i:Int = 1 To f.num

						f.dirx[i] = Sin(i * (360.0 / f.num)) * f.speed[i]
						f.diry[i] = Cos(i * (360.0 / f.num)) * f.speed[i]

						f.x[i]:+f.dirx[i]
						f.y[i]:-f.diry[i]

						Local a:Float = (f.lifetimer + f.life - MilliSecs()) * 1.0 / (f.life / 4.0)

						SetAlpha(a)

						Local col:Int = Rand(128, 255)
						SetColor(col, col, col)

						If f.x[i] > 0 And f.x[i] < SW And f.y[i] > 0 Then

							particlecnt:+1
							DrawImageRect(particle[f.particle[i]], f.x[i], f.y[i], f.size[i], f.size[i])

						EndIf

					Next

				EndIf

				If f.exploded = 1 And MilliSecs() > f.lifetimer + f.life Then

					f.remove
					NewFirework()
					'fireworkcnt:-1

				EndIf

				SetAlpha(1.0)
				SetColor(255, 255, 255)
				SetBlend(ALPHABLEND)

				If f.exploded = 0 And MilliSecs() > f.starttimer

					If f.init = 0 Then
						SetChannelVolume(PlaySound(fly), Rnd(0.25, 1))
						f.init = 1
					EndIf

					Local oldx:Float = f.x[1]
					Local oldy:Float = f.y[1]

					f.x[1]:+f.dirx[1]
					f.y[1]:-f.diry[1]
					f.dirx[1] = f.dirx[1] * 1.05

					Local vecx:Float = f.x[1] - oldx
					Local vecy:Float = f.y[1] - oldy

					SetRotation(ATan2(vecy, vecx) + 90)
					SetAlpha(Normalize(f.y[1], SH, 0, 1, 0))
					DrawImageRect(rocket, f.x[1], f.y[1], 8, 32)

				Else If f.exploded = 0 And MilliSecs() < f.starttimer

					SetRotation(0)
					DrawImageRect(rocket, f.x[1], f.y[1], 8, 32)

				EndIf

			Next

			SetRotation(0)
			SetAlpha(1.0)
			SetColor(255, 255, 255)

			For Local x:Int = -6 To 6

				DrawImageRect(bottle, SW / 2 - 16 + (x * 64), SH * 0.9, 32, 32)

			Next

			SetGadgetText(window, fireworkcnt + " Rockets and " + particlecnt + " particles flying around")

		EndIf

		Flip - 1

	EndIf

	If EventID() = EVENT_GADGETACTION And EventData() = 0 Then End
	
	If EventID() = EVENT_GADGETACTION And EventData() = 1 Then
	
		mode = 1 - mode
		
		If mode = 0 Then
		
			If musicchannel Then StopChannel(musicchannel)
			If music1 Then musicchannel = PlaySound(music1)
			
			Else
		
				If musicchannel Then StopChannel(musicchannel)
			If music2 Then musicchannel = PlaySound(music2)
			
		EndIf
				
	EndIf

	Select EventID()

		Case EVENT_WINDOWCLOSE
			End

		Case EVENT_APPTERMINATE
			End

		Case EVENT_MOUSEDOWN
			If EventData() = 1 Then
		
				NewFirework()
				fireworkcnt:+1
								
			Else
							
				For Local f:firework = EachIn firework.List

					f.remove
					fireworkcnt = 0

				Next
			
				InitFirework()

			EndIf

		Case EVENT_WINDOWSIZE
			SW = ClientWidth(window)
			SH = ClientHeight(window)
			image1 = ResizePixmap(logo1, SW, SH)
			image2 = ResizePixmap(logo2, SW, SH)
			SetGadgetShape(canvas, 0, 0, ClientWidth(window), ClientHeight(window))
			SetViewport 0, 0, ClientWidth(window), ClientHeight(window)

			For s:snowflake = EachIn snowflake.List

				ListRemove(snowflake.List, s)

			Next

			SetGraphics CanvasGraphics(canvas)
			RedrawGadget canvas

	End Select

	If init = True Then
	
		init = False
		If mode = 0 And music1 Then musicchannel = PlaySound(music1)
		If mode = 1 And music2 Then musicchannel = PlaySound(music2)
	
	End If
	
Wend

Function InitFirework()

	' initial first 13 rockets
	For Local i:Int = -6 To 6
	
		NewFirework(i)
		fireworkcnt:+1
		
	Next
	
	' add additional rockets
	For Local i:Int = 1 To rockets - 13
		
		NewFirework()
		fireworkcnt:+1
	
	Next

End Function

Function NewFirework(init:Int = -1000)

	Local x:Float

	Local f:firework = New firework

	If init = -1000 Then

		x = SW / 2 - 5 + (Rand(-6, 6) * 64)

	Else

		x = SW / 2 - 5 + (init * 64)

	EndIf

	Local y:Float = SH * 0.9 - 24
	Local dirx:Float = Rnd(-0.2, 0.2)
	Local diry:Float = Rnd(5, 10)
	
	f.num = Rand(100, 500)

	f.x[1] = x
	f.y[1] = y
	f.dirx[1] = dirx
	f.diry[1] = diry
		
	f.timer = Rnd(0.5, 1.0)
	f.life = Rand(1000, 3000)
	f.lifetimer = 0
	f.exploded = 0
	f.starttimer = MilliSecs() + Rand(500, 2000)

End Function

' ----------------------------------------------------------------------------
' recycle casted snowflake
' ----------------------------------------------------------------------------
Function RecycleFlake(s:snowflake, init:Int = False)
	
	s.x = Rnd(0, SW - 1)
	s.y = 0
	If init Then s.y = Rnd(-SH, 0)

	s.col = Rand(232, 255)
	s.dirx = Rnd(-1, 1)
	s.diry = Rnd(0.5, 2)
	s.style = Rand(0, 3)
	s.sinus = Rnd(360)
	s.alpha = Rnd(0.5, 1)
	
End Function

' ----------------------------------------------------------------------------
' initial cast snowflakes
' ----------------------------------------------------------------------------
Function CreateSnow(flakes:Int = 1)
		
	For Local i:Int = 1 To flakes
		
		CreateNewFlake(True)
		
	Next
	
End Function

' ----------------------------------------------------------------------------
' cast a new snowflake
' ----------------------------------------------------------------------------
Function CreateNewFlake(init:Int = False)
	
	Local s:snowflake = New snowflake
	
	RecycleFlake(s, init)
		
End Function

' ----------------------------------------------------------------------------
' calculate ARGB value
' ----------------------------------------------------------------------------
Function toARGB:Int(a:Int, r:Int, g:Int, b:Int)

	Return ((a Shl 24) | (r Shl 16) | (g Shl 8) | b)
	
End Function

' ----------------------------------------------------------------------------
' read transparent pixels from image for collision
' ----------------------------------------------------------------------------
Function CheckCollision:Int(x:Int, y:Int)

	Return (ReadPixel(image1, x, y) & $ff000000) Shr 24

End Function

' ----------------------------------------------------------------------------
' write pixel to image
' ----------------------------------------------------------------------------
Function UpdateCollision(s:snowflake, x:Int, y:Int, col:Int, alpha:Int)

	Local rgb1:Int, rgb2:Int
	Local r1:Int, g1:Int, b1:Int
	Local r2:Int, g2:Int, b2:Int

	Local rgb:Int
	Local r:Int, g:Int, b:Int
	Local f:String
	
	' random brightness
	rgb1 = ReadPixel(image1, x, y)
	rgb2 = toARGB(col * alpha, col * Rnd(0.9, 1), col * Rnd(0.95, 1), col)
			
	r1 = GetR(rgb1)
	g1 = GetG(rgb1)
	b1 = GetB(rgb1)

	r2 = GetR(rgb2)
	g2 = GetG(rgb2)
	b2 = GetB(rgb2)

	f = "Softlight1"
	r = Filter(r1, b2, f)
	g = Filter(g1, g2, f)
	b = Filter(b1, b2, f)
				
	rgb = toARGB(col * alpha, r, g, b)

	WritePixel(image1, x, y, rgb)

End Function

' ----------------------------------------------------------------------------
' update snowflakes
' ----------------------------------------------------------------------------
Function UpdateSnow:Int(s:snowflake)
	
	Local counter:Int = 0
	Local dist:Int
	Local fx:Int, fy:Int
	Local c:Int, i:Int
	Local ax:Int, ay:Int
	
	Local rgb1:Int, rgb2:Int
	Local r1:Int, g1:Int, b1:Int
	Local r2:Int, g2:Int, b2:Int

	Local rgb:Int
	Local r:Int, g:Int, b:Int
	Local f:String

	SetBlend(ALPHABLEND)
	
	For s:snowflake = EachIn snowflake.List
	
		counter = counter + 1
		
		' smooth random movement
		s.sinus = s.sinus + Rnd(3) Mod 360

		' move horizontal and vertical		
		s.x = s.x + s.dirx + (Sin(s.sinus) * 0.5)
		s.y = s.y + s.diry

		' flake off screen? recycle!		
		If s.y > SH - 1 Then RecycleFlake(s)
		If s.x < 0 Then RecycleFlake(s)
		If s.x > SW Then RecycleFlake(s)
		
		fx = Int(s.x)
		fy = Int(s.y)
			
		' limit horizontal/vertical
		If fx < 1 Then fx = 1 Else If fx > SW - 2 Then fx = SW - 2
			If fy < 1 Then fy = 1 Else If fy > SH - 2 Then fy = SH - 2

				' check for collision with matte				
				If CheckCollision(fx, fy) > 0 And Rnd(1) > chancestick Then
				
					' create new collision pixel in image
					'UpdateCollision(s, fx, fy, s.col, s.alpha)

					' random brightness
					rgb1 = ReadPixel(image1, fx, fy)
					rgb2 = toARGB(s.col * s.alpha, s.col * Rnd(0.9, 1), s.col * Rnd(0.95, 1), s.col)
							
					r1 = GetR(rgb1)
					g1 = GetG(rgb1)
					b1 = GetB(rgb1)
				
					r2 = GetR(rgb2)
					g2 = GetG(rgb2)
					b2 = GetB(rgb2)
				
					f = "Softlight1"
					r = Filter(r1, b2, f)
					g = Filter(g1, g2, f)
					b = Filter(b1, b2, f)
								
					rgb = toARGB(s.col * s.alpha, r, g, b)
					WritePixel(image1, fx, fy, rgb)

					' add aditional pixel
					If Rnd(1) > 0.5 Then
						ax:Int = Rand(-1, 1)
						ay:Int = Rand(-1, 0)
						'UpdateCollision(s, fx + ax, fy + ay, s.col, s.alpha)
						WritePixel(image1, fx + ax, fy + ay, rgb)
					EndIf

					' add aditional pixel
					If Rnd(1) > 0.5 Then
						ax:Int = Rand(-1, 1)
						ay:Int = Rand(-1, 0)
						'UpdateCollision(s, fx + ax, fy + ay, s.col, s.alpha)
						WritePixel(image1, fx + ax, fy + ay, rgb)
					EndIf
						
					' recycle old flake
					RecycleFlake(s)

				Else

					If s.y > 0 Then
						' draw moving flake
						SetAlpha(s.alpha)
						SetColor(255, 255, 255)
						DrawSubImageRect(flake, fx - 1, fy - 1, 3, 3, (s.style * 3), 0, 3, 3)
					EndIf
					
				EndIf
		
				Next

				' cast new snowflakes	
				c = Rand(1, chanceflakes)
				If counter + c <= totalflakes And Rnd(1) > (1 - (snowchance / 100.0)) Then
					For i = 1 To c
						CreateNewFlake()
					Next
				EndIf
	
				Return counter
	
		End Function

		Function Filter:Int(a:Int, b:Int, f:String = "Average")

			Local c:Int

			Select f

				Case "Add" Return Min(255, (a + b))
				Case "Subtract" If (a + b) < 255 Then Return 0 Else Return (a + b) - 255
				Case "Average" Return (a + b) Shr 1
				Case "Multiply" Return (a * b) Shr 8
				Case "Overlay" If a < 128 Then Return (a * b) Shr 7 Else Return 255 - ((255 - a) * (255 - b) Shr 7)
				Case "Exclusion" Return a + b - (a * b Shr 7)
				Case "Negation" Return 255 - Abs(255 - a - b)
				Case "Difference" Return Abs(a - b)
				Case "Darken" If a < b Then Return a Else Return b
				Case "Lighten" If a > b Then Return a Else Return b
				Case "Screen" Return 255 - ((255 - a) * (255 - b) Shr 8)
				Case "Hardlight" If b < 128 Then Return (a * b) Shr 7 Else Return 255 - ((255 - b) * (255 - a) Shr 7)
				Case "Lineardodge" Return Min(a + b, 255)
				Case "Linearburn" Return (a + b) - 192
				Case "Phoenix" Return Min(a, b) - Max(a, b) + 255

				Case "Hardmix" If Filter(a, b, "VividLight") < 128 Then Return 0 Else Return 255
				Case "Vividlight" If a < 128 Then Return Filter(b, (2 * a), "Colorburn") Else Return Filter(b, (2 * (a - 128)), "Colordodge")
				Case "Pinlight" If a < 128 Then Return Filter(b, (2 * a), "Darken") Else Return Filter(b, (2 * (a - 128)), "Lighten")
				Case "Linearlight" If a < 128 Then Return Filter(b, (2 * a), "Linearburn") Else Return Filter(b, (2 * (a - 128)), "Lineardodge")

				Case "Softlight1" c = a * b Shr 8
					Return (c + a * (255 - ((255 - a) * (255 - b) Shr 8) - c) Shr 8)

				Case "Softlight2" c = a * b Shr 8
					Return c + a * (255 - ((255 - a) * (255 - b) Shr 8) - c) Shr 8

				Case "Colordodge" If b = 255 Then
						Return 255
					Else
						c = Floor((a Shl 8) / (255 - b))
						If c > 255 Then Return 255 Else Return c
					EndIf

				Case "Colorburn" If b = 0 Then c = 0 Else c = 255 - (((255 - a) Shl 8) / b)
					If c < 0 Then Return 0 Else Return c

				Case "Glow" If a = 255 Then
						c = 255
					Else
						c = b * b / (255 - a)
						If c > 255 Then Return 255
					EndIf

	
					Return c

			End Select

		End Function

		' ----------------------------------------------------------------------------
		' Return R value of a given RGB value
		' ----------------------------------------------------------------------------
		Function GetR:Int(RGB:Int)
	
			Return RGB Shr 16 & %11111111
	
		End Function

		' ----------------------------------------------------------------------------
		' Return G value of a given RGB value
		' ----------------------------------------------------------------------------
		Function GetG:Int(RGB:Int)
	
			Return RGB Shr 8 & %11111111
	
		End Function

		' ----------------------------------------------------------------------------
		' Return B value of a given RGB value
		' ----------------------------------------------------------------------------
		Function GetB:Int(RGB:Int)
	
			Return RGB & %11111111
	
		End Function

		' ------------------------------------------------------------------------------------------------
		' Normalizes a value to given range
		' ------------------------------------------------------------------------------------------------
		Function Normalize:Float(value:Float, vmin:Float, vmax:Float, nmin:Float, nmax:Float, limit:Int = False)

			' normalize	
			Local result:Float = ((value - vmin) / (vmax - vmin)) * (nmax - nmin) + nmin

			' limit
			If limit Then
	
				If value > nmax Then result = nmax Else If value < nmin Then result = nmin

			EndIf

			Return result
	
		End Function